// edited by hand to make it properly portable
/* include/config.h.  Generated from config.h.in by configure.  */
/* include/config.h.in.  Generated from configure.ac by autoheader.  */

#define HAVE_INT32_T 1
#define HAVE_INT64_T 1
#define HAVE_INTTYPES_H 1
#define HAVE_STDDEF_H 1
#define HAVE_STDINT_H 1
#define HAVE_STDLIB_H 1
#define HAVE_STRING_H 1
#define HAVE_UINT8_T 1
#define INLINE inline
#define PACKAGE_VERSION "2.0.1"
#define PROJECT_VERSION_FULL PACKAGE_VERSION
#define STDC_HEADERS 1
